from libc.stdint cimport uint32_t


cdef extern from "../cxx/cffCompressor.h":
    uint32_t* _compreff "compreff" (
        unsigned char* dataStream, int numRounds,
        unsigned& outputLength) except +
    void unload(uint32_t* response)


def compreff(bytes dataStream, int numRounds):
    cdef unsigned outputLength = 0
    cdef uint32_t* raw_output = _compreff(dataStream, numRounds, outputLength)
    cdef list output = []
    cdef unsigned i
    for i in range(outputLength):
        output.append(raw_output[i])
    if raw_output != NULL:
        unload(raw_output)
    return output



#!/usr/bin/env python
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This script generates a data source file for the C++ compressor.

Usage:
>>> ./gen_data.py /path/to/font.otf /path/to/output
"""

import array
import subprocess
import sys
from fontTools.ttLib import TTFont

from compreffor.cxxCompressor import write_data

if __name__ == '__main__':
    if len(sys.argv) < 3:
        print "missing arguments"
    else:
        f = TTFont(sys.argv[1])
        td = f['CFF '].cff.topDictIndex[0]
        with open(sys.argv[2], 'w') as out_f:
            data = write_data(td)
            out_f.write(data)



#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
==== TTX/FontTools Compreffor ====
This module automatically subroutines the CFF table in
a TTFont object, for the purposes of compressing the
outputted font file. In addition to providing a Python
interface, this tool can be used on the command line.

Usage (python):
>> from fontTools.ttLib import TTFont
>> import compreffor
>> font = TTFont(filename)
>> options = { ... }
>> compreffor.compress(font, **options)
>> font.save(filename)

Compression Backends:
There are 2 different ways the compreffor can be run.
    - The default method is backed by a C++ extension module. The logic is
      in cxxCompressor.py, cffCompressor.h, cffCompressor.cc and
      _compreffor.pyx.
    - The second is a pure Python approach, and can be selected from `compress`
      by passing `method_python=True`. This is significantly slower than the
      the other backend (~10-20x). The logic can be found in pyCompressor.py.

Options:
When running `compreffor.compress`, options can be set using keyword arguments:
    - nrounds (integer) -- the number of market iterations to run (default: 4)
    - max_subrs (integer) -- limit to number of subrs per INDEX
                             (default: 65533)

With `method_python=True`, the following additional options are available:
    - chunk_ratio (float) -- set the percentage of charstrings
                             to be run by each process. The value must be a
                             float between 0 < n <= 1 (default: 0.1)
    - processes (integer) -- the number of simultaneous processes to run.
                             Use value 1 to perform operation serially.

Usage (command line):
From the command line, you can either run the package as a module,

$ python -m compreffor --help

Or call the `compreffor` console script installed with the package.
Use -h/--help to list all the available options.
"""

import logging
from fontTools.misc.loggingTools import Timer

log = logging.getLogger(__name__)
timer = Timer(logger=logging.getLogger(log.name + ".timer"))

from compreffor import cxxCompressor, pyCompressor


try:
    from ._version import version as __version__  # type: ignore
except ImportError:
    __version__ = "0.0.0+unknown"


def compress(ttFont, method_python=False, **options):
    """ Subroutinize TTFont instance in-place using the C++ Compreffor.
    If 'method_python' is True, use the slower, pure-Python Compreffor.
    If the font already contains subroutines, it is first decompressed.
    """
    if has_subrs(ttFont):
        log.warning(
            "There are subroutines in font; must decompress it first")
        decompress(ttFont)
    if method_python:
        pyCompressor.compreff(ttFont, **options)
    else:
        cxxCompressor.compreff(ttFont, **options)


def decompress(ttFont, **kwargs):
    """ Use the FontTools Subsetter to desubroutinize the font's CFF table.
    Any keyword arguments are passed on as options to the Subsetter.
    Skip if the font contains no subroutines.
    """
    if not has_subrs(ttFont):
        log.debug('No subroutines found; skip decompress')
        return

    from fontTools import subset

    # The FontTools subsetter modifies many tables by default; here
    # we only want to desubroutinize, so we run the subsetter on a
    # temporary copy and extract the resulting CFF table from it
    make_temp = kwargs.pop('make_temp', True)
    if make_temp:
        from io import BytesIO
        from fontTools.ttLib import TTFont, newTable

        stream = BytesIO()
        ttFont.save(stream, reorderTables=None)
        stream.flush()
        stream.seek(0)
        tmpfont = TTFont(stream)
    else:
        tmpfont = ttFont  # run subsetter on the original font

    options = subset.Options(**kwargs)
    options.desubroutinize = True
    options.notdef_outline = True
    subsetter = subset.Subsetter(options=options)
    subsetter.populate(glyphs=tmpfont.getGlyphOrder())
    subsetter.subset(tmpfont)

    if make_temp:
        # copy modified CFF table to original font
        data = tmpfont['CFF '].compile(tmpfont)
        table = newTable('CFF ')
        table.decompile(data, ttFont)
        ttFont['CFF '] = table
        tmpfont.close()


def has_subrs(ttFont):
    """ Return True if the font's CFF table contains any subroutines. """
    if 'CFF ' not in ttFont:
        raise ValueError("Invalid font: no 'CFF ' table found")
    td = ttFont['CFF '].cff.topDictIndex[0]
    all_subrs = [td.GlobalSubrs]
    if hasattr(td, 'FDArray'):
        all_subrs.extend(fd.Private.Subrs for fd in td.FDArray
                         if hasattr(fd.Private, 'Subrs'))
    elif hasattr(td.Private, 'Subrs'):
        all_subrs.append(td.Private.Subrs)
    return any(all_subrs)


def check(original_file, compressed_file):
    """ Compare the original and compressed font files to confirm they are
    functionally equivalent. Also check that the Charstrings in the compressed
    font's CFFFontSet don't exceed the maximum subroutine nesting level.
    Return True if all checks pass, else return False.
    """
    from compreffor.test.util import check_compression_integrity
    from compreffor.test.util import check_call_depth
    rv = check_compression_integrity(original_file, compressed_file)
    rv &= check_call_depth(compressed_file)
    return rv


# The `Methods` and `Compreffor` classes are now deprecated, but we keep
# them here for backward compatibility


class Methods:
    Py, Cxx = range(2)


class Compreffor(object):
    def __init__(self, font, method=Methods.Cxx, **options):
        import warnings
        warnings.warn("'Compreffor' class is deprecated; use 'compress' function "
                      "instead", UserWarning)
        self.font = font
        self.method = method
        self.options = options

    def compress(self):
        if self.method == Methods.Py:
            compress(self.font, method_python=True, **self.options)
        elif self.method == Methods.Cxx:
            compress(self.font, method_python=False, **self.options)
        else:
            raise ValueError("Invalid method: %r" % self.method)



#!/usr/bin/env python
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tool to subroutinize a CFF OpenType font. Backed by a C++ binary.

This file is a bootstrap for the C++ edition of the FontTools compreffor.
It prepares the input data for the extension and reads back in the results,
applying them to the input font.

Usage (command line):
>> ./cxxCompressor.py /path/to/font.otf
# font written to /path/to/font.compressed.otf

Usage (python):
>> font = TTFont("/path/to/font.otf")
>> cxxCompressor.compreff(font)
>> font.save("/path/to/output.otf")
"""

import array
from io import BytesIO
import struct
import logging
from compreffor.pyCompressor import (
    Compreffor, CandidateSubr, tokenCost)
from compreffor import _compreffor as lib, timer


log = logging.getLogger(__name__)

__all__ = ["compreff"]


class IdKeyMap(object):
    """A map that where every key's value is itself. Used
    as a map from simplified key space to actual key space
    in pyCompressor"""

    def __getitem__(self, tok):
        return tok


class SimpleCandidateSubr(CandidateSubr):
    """A reimplimentation of CandidateSubr to be more
    compatible with results from C++"""

    def __init__(self, length, ref_loc):
        self.length = length
        self.location = ref_loc
        self.freq = 0
        self._flatten = False
        self._global = False

    def usages(self):
        return self.freq

    frequency = usages

    def cost(self):
        try:
            return self.__cost
        except AttributeError:
            self.__cost = sum(map(tokenCost, self.value()))
            return self.__cost

    def encoding(self):
        return self._encoding


@timer("produce data for C++ library")
def write_data(td):
    """Writes CharStrings from the TopDict td into a string that is easily
    readable."""

    out = BytesIO()
    td.CharStrings.charStringsIndex.getCompiler(td.strings, None).toFile(out)
    return out.getvalue()


def get_encoding(data_buffer, subrs):
    """Read a charstring's encoding stream out of a string buffer response
    from cffCompressor.cc"""

    pos = 0
    num_calls = data_buffer[pos]
    pos += 1
    enc = []
    for j in range(num_calls):
        insertion_pos = struct.unpack_from('<I', data_buffer[pos:pos+4])[0]
        pos += 4
        subr_index = struct.unpack_from('<I', data_buffer[pos:pos+4])[0]
        pos += 4
        subrs[subr_index].freq += 1
        enc.append((insertion_pos, subrs[subr_index]))
    return enc, pos


def read_data(td, result_string):
    """Read the output of cffCompressor.cc into Python data
    structures."""

    results = array.array("B", result_string)
    num_subrs = struct.unpack_from('<I', results[:4])[0]

    # process subrs
    subrs = []
    pos = 4
    for i in range(num_subrs):
        glyph_idx = struct.unpack_from('<I', results[pos:pos+4])[0]
        pos += 4
        tok_idx = struct.unpack_from('<I', results[pos:pos+4])[0]
        pos += 4
        subr_len = struct.unpack_from('<I', results[pos:pos+4])[0]
        pos += 4
        subrs.append(SimpleCandidateSubr(subr_len, (glyph_idx, tok_idx)))
    for i in range(num_subrs):
        enc, num_read = get_encoding(results[pos:], subrs)
        pos += num_read
        subrs[i]._encoding = enc

    # process glyph encodings
    glyph_encodings = []
    for i in range(len(td.CharStrings)):
        enc, num_read = get_encoding(results[pos:], subrs)
        pos += num_read
        glyph_encodings.append(enc)

    assert pos == len(results)
    return (subrs, glyph_encodings)


@timer("extract results")
def interpret_data(td, results):
    """Interpret the result array from a lib.compreff call to
    produce Python data structures."""

    class MutableSpace: pass
    MutableSpace.pos = 0
    def pop_result():
        ans = results[MutableSpace.pos]
        MutableSpace.pos += 1
        return ans

    num_subrs = pop_result()

    # process subrs
    subrs = []
    for i in range(num_subrs):
        glyph_idx = pop_result()
        tok_idx = pop_result()
        subr_len = pop_result()
        subrs.append(SimpleCandidateSubr(subr_len, (glyph_idx, tok_idx)))

    def pop_encoding():
        num_calls = pop_result()
        enc = []
        for j in range(num_calls):
            insertion_pos = pop_result()
            subr_index = pop_result()
            subrs[subr_index].freq += 1
            enc.append((insertion_pos, subrs[subr_index]))
        return enc

    for i in range(num_subrs):
        enc = pop_encoding()
        subrs[i]._encoding = enc

    # process glyph encodings
    glyph_encodings = []
    for i in range(len(td.CharStrings)):
        enc = pop_encoding()
        glyph_encodings.append(enc)

    return (subrs, glyph_encodings)


@timer("compress the font")
def compreff(font, nrounds=None, max_subrs=None):
    """Main function that compresses `font`, a TTFont object,
    in place.
    """
    assert len(font['CFF '].cff.topDictIndex) == 1

    td = font['CFF '].cff.topDictIndex[0]

    if nrounds is None:
        nrounds = Compreffor.NROUNDS
    if max_subrs is None:
        max_subrs = Compreffor.NSUBRS_LIMIT

    input_data = write_data(td)
    with timer("run 'lib.compreff()'"):
        results = lib.compreff(input_data, nrounds)
    subrs, glyph_encodings = interpret_data(td, results)

    with timer("decompile charstrings"):
        for cs in td.CharStrings.values():
            cs.decompile()

    # in order of charset
    chstrings = [x.program for x in td.CharStrings.values()]
    for cs in chstrings:
        Compreffor.collapse_hintmask(cs)

    for s in subrs:
        s.chstrings = chstrings

    if hasattr(td, 'FDSelect'):
        fdselect = lambda g: td.CharStrings.getItemAndSelector(g)[1]
        fdlen = len(td.FDArray)
    else:
        fdselect = None
        fdlen = 1

    nest_limit = Compreffor.SUBR_NEST_LIMIT
    gsubrs, lsubrs = Compreffor.process_subrs(
                            td.charset,
                            glyph_encodings,
                            fdlen,
                            fdselect,
                            subrs,
                            IdKeyMap(),
                            max_subrs,
                            nest_limit)

    encoding = dict(zip(td.charset, glyph_encodings))

    Compreffor.apply_subrs(td, encoding, gsubrs, lsubrs)



#!/usr/bin/env python
#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os
import sys
import time
import numpy as np
from subprocess import call

COMMAND_TEMPLATE = "python ../compreffor/pyCompressor.py /path/to/font.otf --chunkratio %f"

def time_run(value):
    command = COMMAND_TEMPLATE % value

    null = open(os.devnull, "wb")

    start_time = time.time()
    call(command.split(), stdout=null)
    run_time = time.time() - start_time

    return run_time

def minimize_runtime(start_val, stop_val, samples, passes):
    left = start_val
    right = stop_val
    step = float(right - left) / samples

    for i in range(passes):
        print "Testing range (%f, %f) with %f steps" % (left, right, step)
        values = np.arange(left, right + step, step)
        times = map(time_run, values)

        lowest = min(enumerate(times), key=lambda x: x[1])
        low_val, low_time = values[lowest[0]], lowest[1]
        print "Current lowest: %f with %gs" % (low_val, low_time)
        left = low_val - 2 * step
        while left <= 0: left += step 
        right = low_val + 2 * step
        step = float(right - left) / samples

    return low_val

def plot_values(start_val, stop_val, step):
    from matplotlib import pyplot

    values = np.arange(start_val, stop_val + step, step)
    times = map(time_run, values)
    pyplot.plot(values, times)
    pyplot.title("Time to run vs. Changing parameter")
    pyplot.xlabel("Parameter Value")
    pyplot.ylabel("Time to run (seconds)")
    pyplot.show()


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Minimize runtime.')
    parser.add_argument('-p', help='Plot values', action='store_true', default=False,
                        required=False, dest="plot")
    parser.add_argument('start_val', help='What value to start at', type=float)
    parser.add_argument('stop_val', help='What value to stop at', type=float)
    parser.add_argument('step', help='What value to step by (for plotter) or number of samples (for minimizer)', type=float)
    parser.add_argument('passes', help='How many passes to run (for minimizer)', type=int,
                        nargs="?")

    args = parser.parse_args()

    if not args.plot:
        assert args.passes != None, "passes argument required"
        min_val = minimize_runtime(args.start_val, args.stop_val, args.step, args.passes)
        print "Minimized time value: %f" % min_val
    else:
        plot_values(args.start_val, args.stop_val, args.step)



#
# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import logging
from fontTools import ttLib
from fontTools.misc import psCharStrings
from compreffor import decompress, timer


log = logging.getLogger(__name__)


@timer("check compression integrity")
def check_compression_integrity(orignal_file, compressed_file):
    """Compares two fonts to confirm they are functionally equivalent"""

    orig_font = ttLib.TTFont(orignal_file)
    orig_gset = orig_font.getGlyphSet()
    comp_font = ttLib.TTFont(compressed_file)
    comp_gset = comp_font.getGlyphSet()

    assert orig_gset.keys() == comp_gset.keys()

    decompress(orig_font, make_temp=False)
    decompress(comp_font, make_temp=False)

    passed = True
    for g in orig_gset.keys():
        orig_glyph = orig_gset[g]._glyph
        comp_glyph = comp_gset[g]._glyph
        orig_glyph.decompile()
        if not (orig_glyph.program == comp_glyph.program):
            log.warning("Difference found in glyph '%s'" % (g,))
            passed = False

    if passed:
        log.info("Fonts match!")
        return True
    else:
        log.warning("Fonts have differences :(")
        return False


@timer("check subroutine nesting depth")
def check_call_depth(compressed_file):
    """Runs `check_cff_call_depth` on a file"""

    f = ttLib.TTFont(compressed_file)

    return check_cff_call_depth(f["CFF "].cff)


def check_cff_call_depth(cff):
    """Checks that the Charstrings in the provided CFFFontSet
    obey the rules for subroutine nesting.

    Return True if the subroutine nesting level does not exceed
    the maximum limit (10), else return False.
    """

    SUBR_NESTING_LIMIT = 10

    assert len(cff.topDictIndex) == 1

    td = cff.topDictIndex[0]

    class track_info: pass

    track_info.max_for_all = 0

    gsubrs = cff.GlobalSubrs
    gbias = psCharStrings.calcSubrBias(gsubrs)

    def follow_program(program, depth, subrs):
        bias = psCharStrings.calcSubrBias(subrs)

        if len(program) > 0:
            last = program[0]
            for tok in program[1:]:
                if tok == "callsubr":
                    assert type(last) == int
                    next_subr = subrs[last + bias]
                    if (not hasattr(next_subr, "_max_call_depth") or
                            next_subr._max_call_depth < depth + 1):
                        increment_subr_depth(next_subr, depth + 1, subrs)
                elif tok == "callgsubr":
                    assert type(last) == int
                    next_subr = gsubrs[last + gbias]
                    if (not hasattr(next_subr, "_max_call_depth") or
                            next_subr._max_call_depth < depth + 1):
                        increment_subr_depth(next_subr, depth + 1, subrs)
                last = tok
        else:
            log.warning("Compiled subr encountered")

    def increment_subr_depth(subr, depth, subrs=None):
        if not hasattr(subr, "_max_call_depth") or subr._max_call_depth < depth:
            subr._max_call_depth = depth

        if subr._max_call_depth > track_info.max_for_all:
            track_info.max_for_all = subr._max_call_depth

        program = subr.program
        follow_program(program, depth, subrs)

    for cs in td.CharStrings.values():
        cs.decompile()
        follow_program(cs.program, 0, getattr(cs.private, "Subrs", []))

    if track_info.max_for_all <= SUBR_NESTING_LIMIT:
        log.info("Subroutine nesting depth ok! [max nesting depth of %d]",
                 track_info.max_for_all)
        return True
    else:
        log.warning("Subroutine nesting depth too deep :( [max nesting depth "
                    "of %d]", track_info.max_for_all)
        return False
